'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var EventEmitter = require('events');
var diff = require('fast-diff');
var cloneDeep = require('lodash/cloneDeep');
var cloneDeepWith = require('lodash/cloneDeepWith');
var isEqualWith = require('lodash/isEqualWith');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var EventEmitter__default = /*#__PURE__*/_interopDefaultLegacy(EventEmitter);
var diff__default = /*#__PURE__*/_interopDefaultLegacy(diff);
var cloneDeep__default = /*#__PURE__*/_interopDefaultLegacy(cloneDeep);
var cloneDeepWith__default = /*#__PURE__*/_interopDefaultLegacy(cloneDeepWith);
var isEqualWith__default = /*#__PURE__*/_interopDefaultLegacy(isEqualWith);

// This class correctly handles the counting of unicode code points
// and provides the same behaviour as python strings
// we need this to play nice with the backend
// all standard methods follow the js string specs https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/String
class DeltaString {
  constructor(nativeString = '') {
    let characterArray;

    if (nativeString instanceof Array) {
      characterArray = nativeString;
      nativeString = characterArray.join('');
    }

    if (nativeString instanceof DeltaString) {
      characterArray = Array.from(nativeString.characterArray);
      nativeString = nativeString.toString();
    } else {
      characterArray = Array.from(nativeString);
    }

    Object.defineProperties(this, {
      nativeString: {
        value: nativeString
      },
      characterArray: {
        value: characterArray
      },
      length: {
        value: characterArray.length
      },
      __v_skip: {
        value: true
      }
    });
    return new Proxy(this, {
      get(target, prop, receiver) {
        if (typeof prop === 'string' && Number.isInteger(Number(prop))) {
          return target.characterArray[prop];
        }

        return Reflect.get(...arguments);
      }

    });
  } // extensions
  // computes the correct DeltaString index from a native index


  transformFromNativeIndex(nativeIndex) {
    let index = 0;

    for (const char of this.characterArray) {
      if (nativeIndex <= 0) return index;
      nativeIndex -= char.length;
      index++;
    }

    return index;
  } // computes the correct native index from a deltastring index


  transformToNativeIndex(index) {
    let nativeIndex = 0;

    for (let i = 0; i < index && i < this.characterArray.length; i++) {
      nativeIndex += this.characterArray[i].length;
    }

    return nativeIndex;
  } // works like Array.splice, but clones
  // TODO tests


  splice() {
    // TODO better performance?
    const clone = this.characterArray.slice();
    clone.splice(...arguments);
    return new DeltaString(clone);
  } // standard methods


  [Symbol.iterator]() {
    return this.characterArray[Symbol.iterator]();
  }

  concat() {
    return new DeltaString(this.nativeString.concat(...arguments));
  }

  indexOf(other) {
    const otherArray = other instanceof DeltaString ? other.characterArray : Array.from(other);
    return this.characterArray.findIndex((character, index) => {
      for (let i = 0; i < otherArray.length; i++) {
        if (this.characterArray[index + i] !== otherArray[i]) return false;
      }

      return true;
    });
  }

  slice(beginIndex, endIndex) {
    return new DeltaString(this.characterArray.slice(beginIndex, endIndex));
  }

  substring(indexStart, indexEnd) {
    var _indexEnd;

    indexStart = Math.max(0, Math.min(this.length, indexStart || 0));
    indexEnd = Math.max(0, Math.min(this.length, (_indexEnd = indexEnd) !== null && _indexEnd !== void 0 ? _indexEnd : this.length));
    return indexStart <= indexEnd ? this.slice(indexStart, indexEnd) : this.slice(indexEnd, indexStart);
  }

  substr(start, length) {
    if (length < 0) {
      length = 0;
    }

    if (length !== undefined && start !== 0) {
      length = Math.sign(start) * (Math.abs(start) + length);
    }

    return this.slice(start, length);
  }

  valueOf() {
    return this.nativeString;
  }

  toString() {
    return this.nativeString;
  }

  toJSON() {
    return this.nativeString;
  }

  equals(other) {
    return this.valueOf() == other; // eslint-disable-line eqeqeq
  } // charAt
  // charCodeAt
  // codePointAt
  // includes
  // endsWith
  // lastIndexOf
  // localeCompare
  // match
  // matchAll
  // normalize
  // padEnd
  // padStart
  // repeat
  // replace
  // search
  // split
  // startsWith
  // toLocaleLowerCase
  // toLocaleUpperCase
  // toLowerCase
  // toSource
  // toUpperCase
  // trim
  // trimStart
  // trimLeft
  // trimEnd
  // trimRight


}

const DELTA = Symbol('DELTA');
const DELTA_STR = Symbol('DELTA_STR');
const BASE_TYPES = {
  DELTA,
  DELTA_STR
};
const SUBTYPES = {};
function setSubtypes(newSubtypes) {
  Object.assign(SUBTYPES, newSubtypes);
}

const deltaStringEqualCustomizer = function (a, b) {
  if (a instanceof DeltaString && b instanceof DeltaString) {
    return a.equals(b);
  }

  if (a instanceof DeltaString || b instanceof DeltaString) {
    return a == b; // eslint-disable-line eqeqeq
  }
};

function isEqual(a, b) {
  return isEqualWith__default["default"](a, b, deltaStringEqualCustomizer);
}

const deltaStringCloneCustomizer = function (value) {
  if (value instanceof DeltaString) {
    return value.slice();
  }
};

function clone(obj) {
  return cloneDeepWith__default["default"](obj, deltaStringCloneCustomizer);
}
function getOpLength(op) {
  if (typeof op.delete === 'number') {
    return op.delete;
  } else if (typeof op.retain === 'number') {
    return op.retain;
  } else {
    return op.insert instanceof DeltaString ? op.insert.length : 1;
  }
}
function convertOps(ops) {
  const convertOp = function (op) {
    var _op$$sub, _op$insert;

    // convert strings to DeltaString
    if (typeof op.insert === 'string') {
      op.insert = new DeltaString(op.insert);
      return op;
    }

    if ((_op$$sub = op.$sub) !== null && _op$$sub !== void 0 && _op$$sub.items) {
      convertOps(op.$sub.items);
      return op;
    }

    if (op.$sub) {
      convertOps(op.$sub);
      return op;
    }

    const typeSpec = SUBTYPES[(_op$insert = op.insert) === null || _op$insert === void 0 ? void 0 : _op$insert._t];

    if (typeSpec) {
      for (const [key, value] of Object.entries(op.insert)) {
        if (typeSpec[key] === BASE_TYPES.DELTA_STR && !(value instanceof DeltaString)) {
          op.insert[key] = new DeltaString(value);
        }

        if (typeSpec[key] === BASE_TYPES.DELTA) {
          convertOps(op.insert[key]);
        }
      }
    }
  };

  if (!(ops instanceof Array)) {
    return convertOp(ops);
  }

  for (const op of ops) {
    convertOp(op);
  }

  return ops;
}
const attributes = {
  compose(a, b, keepNull) {
    if (typeof a !== 'object') a = {};
    if (typeof b !== 'object') b = {};
    let attributes = cloneDeep__default["default"](b);

    if (!keepNull) {
      attributes = Object.keys(attributes).reduce(function (copy, key) {
        if (attributes[key] != null) {
          copy[key] = attributes[key];
        }

        return copy;
      }, {});
    } // TODO make deep?


    for (const key in a) {
      if (a[key] !== undefined && b[key] === undefined) {
        attributes[key] = a[key];
      }
    }

    return Object.keys(attributes).length > 0 ? attributes : undefined;
  },

  diff(a, b) {
    if (typeof a !== 'object') a = {};
    if (typeof b !== 'object') b = {};
    const attributes = Object.keys(a).concat(Object.keys(b)).reduce(function (attributes, key) {
      if (!isEqual(a[key], b[key])) {
        attributes[key] = b[key] === undefined ? null : b[key];
      }

      return attributes;
    }, {});
    return Object.keys(attributes).length > 0 ? attributes : undefined;
  },

  transform: function (a, b, priority) {
    if (typeof a !== 'object') return b;
    if (typeof b !== 'object') return undefined;
    if (!priority) return b; // b simply overwrites us without priority

    const attributes = Object.keys(b).reduce(function (attributes, key) {
      if (a[key] === undefined) attributes[key] = b[key]; // null is a valid value

      return attributes;
    }, {});
    return Object.keys(attributes).length > 0 ? attributes : undefined;
  }
};

// lifted from https://github.com/quilljs/delta/blob/master/lib/op.js
class OpIterator {
  constructor(ops) {
    this.ops = ops;
    this.index = 0;
    this.offset = 0;
  } // override for custom behaviour


  getOpLength(op) {
    return getOpLength(op);
  }

  copyInsert(op, offset, length) {
    if (op.insert instanceof DeltaString) {
      return op.insert.substr(offset, length);
    }

    if (typeof op.insert === 'object') {
      return op.insert;
    }
  }

  hasNext() {
    return this.peekLength() < Infinity;
  }

  next(length) {
    if (!length) length = Infinity;
    const nextOp = this.ops[this.index];

    if (!nextOp) {
      return {
        retain: Infinity
      };
    }

    const offset = this.offset;
    const opLength = this.getOpLength(nextOp);

    if (length >= opLength - offset) {
      length = opLength - offset;
      this.index++;
      this.offset = 0;
    } else {
      this.offset += length;
    }

    if (typeof nextOp.delete === 'number') {
      return {
        delete: length
      };
    } else {
      const retOp = {};

      if (nextOp.attributes) {
        retOp.attributes = nextOp.attributes;
      }

      if (typeof nextOp.retain === 'number') {
        retOp.retain = length;

        if (nextOp.$sub) {
          retOp.$sub = nextOp.$sub;
        }

        if (nextOp.$set) {
          retOp.$set = nextOp.$set;
        }
      }

      if (nextOp.insert) {
        retOp.insert = this.copyInsert(nextOp, offset, length);
      }

      return retOp;
    }
  }

  peek() {
    return this.ops[this.index];
  }

  peekLength() {
    if (this.ops[this.index]) {
      return this.getOpLength(this.ops[this.index]) - this.offset;
    }

    return Infinity;
  }

  peekType() {
    if (this.ops[this.index]) {
      if (typeof this.ops[this.index].delete === 'number') {
        return 'delete';
      } else if (typeof this.ops[this.index].retain === 'number') {
        return 'retain';
      } else {
        return 'insert';
      }
    }

    return 'retain';
  }

}

// lifted from https://github.com/quilljs/delta/blob/master/lib/delta.js
const NULL_CHARACTER = String.fromCharCode(0); // Placeholder char for embed in diff()

const composeOplist = function (a, b) {
  if (a && b) {
    return new Delta(a).compose(new Delta(b)).ops;
  } else if (a) {
    return clone(a);
  } else if (b) {
    return clone(b);
  }
};

class Delta {
  constructor(ops = []) {
    convertOps(ops);
    this.ops = ops;
  }

  insert(text, attributes) {
    // TODO canonical op ordering
    if (text.length === 0) return this;
    const op = {
      insert: text
    };

    if (attributes != null && typeof attributes === 'object' && Object.keys(attributes).length > 0) {
      op.attributes = attributes;
    }

    return this.push(op);
  }

  delete(length) {
    if (length <= 0) return this;
    const op = {
      delete: length
    };
    return this.push(op);
  }

  retain(length, args) {
    const {
      attributes,
      subOps,
      set
    } = args !== null && args !== void 0 ? args : {};
    if (length <= 0) return this;
    const op = {
      retain: length
    };

    if (attributes != null && typeof attributes === 'object' && Object.keys(attributes).length > 0) {
      op.attributes = attributes;
    }

    if (set != null && typeof set === 'object' && Object.keys(set).length > 0) {
      op.$set = set;
    } // subOps can be Delta, object, or array


    if (subOps != null && typeof subOps === 'object' && Object.keys(subOps).length > 0) {
      if (subOps instanceof Delta) {
        op.$sub = subOps.ops;
      } else {
        op.$sub = subOps;
      }
    }

    return this.push(op);
  }

  push(newOp) {
    let index = this.ops.length;
    let lastOp = this.ops[index - 1];
    convertOps(newOp); // patch lastOp if we can

    if (typeof lastOp === 'object') {
      if (newOp.delete && lastOp.delete) {
        this.ops[index - 1] = {
          delete: lastOp.delete + newOp.delete
        };
        return this;
      } // Since it does not matter if we insert before or after deleting at the same index,
      // always prefer to insert first


      if (typeof lastOp.delete === 'number' && newOp.insert != null) {
        index -= 1;
        lastOp = this.ops[index - 1];

        if (!lastOp) {
          this.ops.unshift(newOp);
          return this;
        }
      }

      if (isEqual(newOp.attributes, lastOp.attributes)) {
        if (newOp.insert instanceof DeltaString && lastOp.insert instanceof DeltaString) {
          this.ops[index - 1] = {
            insert: lastOp.insert.concat(newOp.insert)
          };
          if (newOp.attributes) this.ops[index - 1].attributes = newOp.attributes;
          return this;
        }

        if (typeof newOp.retain === 'number' && typeof lastOp.retain === 'number' && isEqual(newOp.$sub, lastOp.$sub) && isEqual(newOp.$set, lastOp.$set)) {
          this.ops[index - 1] = {
            retain: lastOp.retain + newOp.retain
          };
          if (newOp.attributes) this.ops[index - 1].attributes = newOp.attributes;
          if (newOp.$sub) this.ops[index - 1].$sub = newOp.$sub;
          if (newOp.$set) this.ops[index - 1].$set = newOp.$set;
          return this;
        }
      }
    }

    if (index === this.ops.length) {
      this.ops.push(newOp);
    } else {
      this.ops.splice(index, 0, newOp);
    }

    return this;
  }

  apply(source) {
    // TODO range checks
    const newString = [];
    let sourceIndex = 0;

    for (const op of this.ops) {
      if (op.retain) {
        newString.push(source.slice(sourceIndex, sourceIndex + op.retain));
        sourceIndex += op.retain;
      } else if (op.insert) {
        newString.push(op.insert);
      } else if (op.delete) {
        sourceIndex += op.delete;
      }
    } // retain the remaining string


    newString.push(source.slice(sourceIndex));
    return newString.join('');
  }

  chop() {
    const lastOp = this.ops[this.ops.length - 1];

    if (lastOp && lastOp.retain && !lastOp.attributes && !lastOp.$sub && !lastOp.$set) {
      this.ops.pop();
    }

    return this;
  }

  map(predicate) {
    return this.ops.map(predicate);
  } // returns a normalized clone (basically throws all ops through delta.push recursivly)


  normalize() {
    const normalizedDelta = new Delta();

    for (let op of this.ops) {
      var _op$insert;

      op = clone(op);

      if ((_op$insert = op.insert) !== null && _op$insert !== void 0 && _op$insert.items) {
        op.insert.items = new Delta(op.insert.items).normalize().ops;
      }

      if (op.$sub) {
        if (op.$sub instanceof Array) {
          op.$sub = new Delta(op.$sub).normalize().ops;
        } else {
          for (const [key, ops] of Object.entries(op.$sub)) {
            op.$sub[key] = new Delta(ops).normalize().ops;
          }
        }
      }

      normalizedDelta.push(op);
    }

    return normalizedDelta;
  } // Compose merges two consecutive operations into one operation, that
  // preserves the changes of both. Or, in other words, for each input string S
  // and a pair of consecutive operations A and B,
  // apply(apply(S, A), B) = apply(S, compose(A, B)) must hold.


  compose(otherDelta) {
    const thisIter = new OpIterator(this.ops);
    const otherIter = new OpIterator(otherDelta.ops);
    const newDelta = new Delta();

    while (thisIter.hasNext() || otherIter.hasNext()) {
      if (otherIter.peekType() === 'insert') {
        // new insert always gets used
        newDelta.push(otherIter.next());
      } else if (thisIter.peekType() === 'delete') {
        // old delete always gets used
        newDelta.push(thisIter.next());
      } else {
        const length = Math.min(thisIter.peekLength(), otherIter.peekLength());
        const thisOp = thisIter.next(length);
        const otherOp = otherIter.next(length);

        if (typeof otherOp.retain === 'number') {
          const newOp = {};

          if (typeof thisOp.retain === 'number') {
            // if both retain, also retain
            newOp.retain = length;

            if (thisOp.$set || otherOp.$set) {
              newOp.$set = attributes.compose(thisOp.$set, otherOp.$set, true);
            }

            if (thisOp.$sub && otherOp.$sub) {
              const thisSubs = thisOp.$sub instanceof Array ? {
                items: thisOp.$sub
              } : thisOp.$sub || {};
              const otherSubs = otherOp.$sub instanceof Array ? {
                items: otherOp.$sub
              } : otherOp.$sub || {};
              const newSubs = Object.assign({}, thisSubs);

              for (const [key, value] of Object.entries(otherSubs)) {
                newSubs[key] = composeOplist(thisSubs[key], value);
              }

              newOp.$sub = newSubs.items && Object.keys(newSubs).length === 1 ? newSubs.items : newSubs;
            } else if (thisOp.$sub || otherOp.$sub) {
              newOp.$sub = composeOplist(thisOp.$sub, otherOp.$sub);
            }

            if (newOp.$sub && !Object.keys(newOp.$sub).length) {
              delete newOp.$sub;
            }
          } else {
            if (thisOp.insert instanceof DeltaString || !otherOp.$sub && !otherOp.$set) {
              newOp.insert = thisOp.insert; // old insert overrides new retain
            } else {
              const typeMark = thisOp.insert._t;
              if (!SUBTYPES[typeMark]) throw new TypeError(`Invalid item type ${typeMark}`);
              const typeSpec = SUBTYPES[typeMark]; // TODO typechecking

              newOp.insert = attributes.compose(thisOp.insert, otherOp.$set, false);

              if (otherOp.$sub) {
                const otherSubs = otherOp.$sub instanceof Array ? {
                  items: otherOp.$sub
                } : otherOp.$sub || {};

                for (const [key, value] of Object.entries(otherSubs)) {
                  if (typeSpec[key] === BASE_TYPES.DELTA_STR) {
                    newOp.insert[key] = new DeltaString(new Delta(value).apply(thisOp.insert[key] || ''));
                  } else if (typeSpec[key] === BASE_TYPES.DELTA) {
                    newOp.insert[key] = composeOplist(thisOp.insert[key], value);
                  } else {
                    throw new TypeError(` ${typeMark}.${key} is not a delta`);
                  }
                }
              }
            }
          } // Preserve null when composing with a retain, otherwise remove it for inserts


          const attributes$1 = attributes.compose(thisOp.attributes, otherOp.attributes, typeof thisOp.retain === 'number');

          if (attributes$1) {
            newOp.attributes = attributes$1;
          }

          newDelta.push(newOp); // new op should be delete, old op is either insert or retain
          // new delete and old insert cancel out, new delete overrides old remain
        } else if (typeof otherOp.delete === 'number' && typeof thisOp.retain === 'number') {
          newDelta.push(otherOp);
        }
      }
    }

    return newDelta.chop();
  }

  diff(otherDelta, index) {
    if (this.ops === otherDelta.ops) {
      return new Delta();
    }

    const strings = [this, otherDelta].map(function (delta) {
      return delta.map(function (op) {
        if (op.insert != null) {
          return op.insert instanceof DeltaString ? op.insert.toString() : NULL_CHARACTER;
        }

        const prep = delta === otherDelta ? 'on' : 'with';
        throw new Error('diff() called ' + prep + ' non-document');
      }).join('');
    });
    const newDelta = new Delta();
    const diffResult = diff__default["default"](strings[0], strings[1], index);
    const thisIter = new OpIterator(this.ops);
    const otherIter = new OpIterator(otherDelta.ops);
    diffResult.forEach(function (component) {
      let length = Array.from(component[1]).length;

      while (length > 0) {
        let opLength = 0;

        switch (component[0]) {
          case diff__default["default"].INSERT:
            opLength = Math.min(otherIter.peekLength(), length);
            newDelta.push(otherIter.next(opLength));
            break;

          case diff__default["default"].DELETE:
            opLength = Math.min(length, thisIter.peekLength());
            thisIter.next(opLength);
            newDelta.delete(opLength);
            break;

          case diff__default["default"].EQUAL:
            {
              opLength = Math.min(thisIter.peekLength(), otherIter.peekLength(), length);
              const thisOp = thisIter.next(opLength);
              const otherOp = otherIter.next(opLength);

              if (isEqual(thisOp.insert, otherOp.insert)) {
                newDelta.retain(opLength, {
                  attributes: attributes.diff(thisOp.attributes, otherOp.attributes)
                });
              } else {
                newDelta.push(otherOp).delete(opLength);
              }

              break;
            }
        }

        length -= opLength;
      }
    });
    return newDelta.chop();
  } // Transform takes two operations A and B that happened concurrently and
  // produces two operations A' and B' (in an array) such that
  // `apply(apply(S, A), B') = apply(apply(S, B), A')`. This function is the
  // heart of OT.


  transform(otherDelta, hasPriority) {
    const thisIter = new OpIterator(this.ops);
    const otherIter = new OpIterator(otherDelta.ops);
    const newDelta = new Delta();

    while (thisIter.hasNext() || otherIter.hasNext()) {
      if (thisIter.peekType() === 'insert' && (hasPriority || otherIter.peekType() !== 'insert')) {
        newDelta.retain(getOpLength(thisIter.next()));
      } else if (otherIter.peekType() === 'insert') {
        newDelta.push(otherIter.next());
      } else {
        const length = Math.min(thisIter.peekLength(), otherIter.peekLength());
        const thisOp = thisIter.next(length);
        const otherOp = otherIter.next(length);

        if (thisOp.delete) {
          // Our delete either makes their delete redundant or removes their retain
          continue;
        } else if (otherOp.delete) {
          newDelta.push(otherOp);
        } else {
          // We retain either their retain or insert
          let newSubs;

          if (thisOp.$sub && otherOp.$sub) {
            const thisSubs = thisOp.$sub instanceof Array ? {
              items: thisOp.$sub
            } : thisOp.$sub || {};
            const otherSubs = otherOp.$sub instanceof Array ? {
              items: otherOp.$sub
            } : otherOp.$sub || {};
            newSubs = {};

            for (const [key, value] of Object.entries(otherSubs)) {
              if (thisSubs[key]) {
                newSubs[key] = new Delta(thisSubs[key]).transform(new Delta(value), hasPriority).ops;
              } else {
                newSubs[key] = clone(value);
              }
            }

            newSubs = newSubs.items && Object.keys(newSubs).length === 1 ? newSubs.items : newSubs;
          } else if (otherOp.$sub) {
            newSubs = clone(otherOp.$sub);
          }

          newDelta.retain(length, {
            attributes: attributes.transform(thisOp.attributes, otherOp.attributes, hasPriority),
            set: attributes.transform(thisOp.$set, otherOp.$set, hasPriority),
            subOps: newSubs
          });
        }
      }
    }

    return newDelta.chop();
  }

  transformPosition(path, priority = false, ops = this.ops) {
    if (!path.length) return path;
    const iter = new OpIterator(ops);
    let offset = 0;
    let index = path[0];

    while (iter.hasNext() && offset <= index) {
      const type = iter.peekType();
      const hasSub = type === 'retain' && iter.peek().$sub;
      const op = iter.next(hasSub ? 1 : Infinity);
      const length = iter.getOpLength(op);

      if (type === 'delete') {
        index -= length;

        if (offset > index) {
          // passed the original position with a delete. If there are deeper levels, we are booted out
          return [offset];
        }

        continue;
      } else if (type === 'insert' && (offset < index || !priority)) {
        index += length;
      } else if (hasSub && offset === index) {
        if (typeof path[1] === 'string') {
          const subOps = op.$sub[path[1]];
          return path.slice(0, 2).concat(this.transformPosition(path.slice(2), priority, subOps));
        } else {
          const subOps = op.$sub instanceof Array ? op.$sub : op.$sub.items || op.$sub.text;
          if (subOps) return [index].concat(this.transformPosition(path.slice(1), priority, subOps));
        }
      }

      offset += length;
    }

    return [index].concat(path.slice(1));
  } // plaintext diffing


  static diff(a, b) {
    const aDelta = new Delta().insert(a);
    const bDelta = new Delta().insert(b);
    return aDelta.diff(bDelta);
  }

}

class APIError extends Error {
  constructor(apiError, ...params) {
    // Pass remaining arguments (including vendor specific ones) to parent constructor
    super(apiError.message || JSON.stringify(apiError.error), ...params); // Maintains proper stack trace for where our error was thrown (only available on V8)

    if (Error.captureStackTrace) {
      Error.captureStackTrace(this, APIError);
    }

    this.name = 'APIError'; // Custom debugging information

    this.apiError = apiError;
  }

}

/* global WebSocket */

const defer = function () {
  const deferred = {};
  deferred.promise = new Promise(function (resolve, reject) {
    deferred.resolve = resolve;
    deferred.reject = reject;
  });
  return deferred;
};

class QuidditchClient extends EventEmitter__default["default"] {
  constructor(url, config) {
    super();
    const defaultConfig = {
      pingInterval: 10000,
      callTimeout: 10000,
      joinTimeout: 60000,
      reconnectDelay: 1000,
      sendSelectInterval: 600,
      token: ''
    };
    this._config = Object.assign(defaultConfig, config);
    this._url = url;
    this._otChannels = {}; // keyed by channel name: {deltaInFlight, buffer, rev}

    this._currentView = null;

    this._createSocket();
  }

  close() {
    this._normalClose = true;

    this._socket.close();

    clearTimeout(this._joinTimeout);
    clearTimeout(this._authenticationTimeout);
    clearTimeout(this._sendSelectTimeout);
    this._sendSelectTimeout = null;
  }

  call(name, data, opts) {
    const options = {
      timeout: this._config.callTimeout
    };
    Object.assign(options, opts);

    const {
      id,
      promise
    } = this._createRequest();

    const payload = [name, id, data];

    this._send(JSON.stringify(payload));

    setTimeout(() => {
      if (this._openRequests[id]) {
        this._timeoutedRequests[id] = true;

        const timeoutedRequest = this._popPendingRequest(id);

        timeoutedRequest.deferred.reject(new Error('call timed out'));
      }
    }, options.timeout);
    return promise;
  }

  sendDelta(channelName, delta) {
    let channel = this._otChannels[channelName];

    if (!channel) {
      channel = this._otChannels[channelName] = {
        deltaInFlight: null,
        buffer: null,
        rev: 0
      };
    }

    const handleAck = function ({
      rev
    }, deferred) {
      channel.deltaInFlight = null;
      channel.rev = rev;

      if (channel.buffer) {
        const buffer = channel.buffer;
        channel.buffer = null;
        sendDelta(buffer.delta).then(() => {
          for (const promise of buffer.promises) {
            promise.resolve();
          }
        }).catch(error => {
          for (const promise of buffer.promises) {
            promise.reject(error);
          }
        });
      }

      deferred.resolve();
    };

    const sendDelta = deltaToSend => {
      const {
        id,
        promise
      } = this._createRequest(handleAck);

      channel.deltaInFlight = deltaToSend;
      const payload = ['ot:delta', id, channelName, {
        delta: deltaToSend.ops,
        rev: channel.rev
      }];

      this._send(JSON.stringify(payload));

      return promise;
    };

    if (!channel.deltaInFlight) {
      return sendDelta(delta);
    } else {
      const deferred = defer();

      if (channel.buffer) {
        channel.buffer.delta = channel.buffer.delta.compose(delta);
        channel.buffer.promises.push(deferred);
      } else {
        channel.buffer = {
          delta,
          promises: [deferred]
        };
      }

      return deferred.promise;
    }
  }

  sendSelect(reference, data) {
    const payload = ['user:select', reference];
    if (data) payload.push(data);

    if (!this._sendSelectTimeout) {
      this._send(JSON.stringify(payload));

      this._sendSelectPayload = null;
      this._sendSelectTimeout = setTimeout(() => {
        this._sendSelectTimeout = null;
        if (this._sendSelectPayload) this._send(JSON.stringify(this._sendSelectPayload));
        this._sendSelectPayload = null;
      }, this._config.sendSelectInterval);
    } else {
      this._sendSelectPayload = payload;
    }
  }

  sendView(reference) {
    const payload = JSON.stringify(['user:view', reference]);
    if (payload === this._currentView) return;
    if (this.socketState === 'open') this._send(payload);
    this._currentView = payload;
  }

  closeChannels(prefix) {
    for (const channelName of Object.keys(this._otChannels)) {
      if (channelName.startsWith(prefix)) this._closeChannel(channelName);
    }
  } // ===========================================================================
  // INTERNALS
  // ===========================================================================


  _createSocket() {
    this._socket = new WebSocket(this._url);
    this.socketState = 'connecting'; // 'closed', 'open', 'connecting'

    this._pingState = {
      latestPong: 0
    };
    this.normalClose = false;

    this._socket.addEventListener('open', () => {
      this.emit('open');
      this.socketState = 'open';

      this._authenticate();

      this._authenticationTimeout = setTimeout(() => {
        this._handlePingTimeout();
      }, this._config.pingInterval);
    });

    this._socket.addEventListener('close', event => {
      this.socketState = 'closed';

      for (const channelName of Object.keys(this._otChannels)) {
        this._closeChannel(channelName);
      }

      this.emit('closed', event); // why past tense? because the socket is already closed and not currently closing

      if (!this._normalClose) {
        setTimeout(() => {
          this.emit('reconnecting');

          this._createSocket();
        }, this._config.reconnectDelay); // throttle reconnect
      }
    });

    this._socket.addEventListener('message', this._processMessage.bind(this));

    this._openRequests = {}; // save deferred promises from requests waiting for reponse

    this._timeoutedRequests = {}; // save timed out request ids so we can drop responses

    this._nextRequestIndex = 1; // autoincremented rohrpost message id

    this._authenticationTimeout = null;
    this._joinTimeout = null;
  }

  _send(payload) {
    this._socket.send(payload);

    this.emit('log', {
      direction: 'send',
      data: payload
    });
  }

  _authenticate() {
    const payload = ['auth', {
      token: this._config.token
    }];

    this._send(JSON.stringify(payload));
  }

  _ping(starterSocket) {
    // we need a ref to the socket to detect reconnects and stop the old ping loop
    if (this._socket.readyState !== 1) return;
    const timestamp = Date.now();
    const payload = ['ping', timestamp];

    this._send(JSON.stringify(payload));

    this.emit('ping');
    setTimeout(() => {
      if (this._socket.readyState !== 1 || this._socket !== starterSocket) return; // looping on old socket, abort

      if (timestamp > this._pingState.latestPong) // we received no pong after the last ping
        this._handlePingTimeout();else this._ping(starterSocket);
    }, this._config.pingInterval);
  }

  _handlePingTimeout() {
    this._socket.close();
  }

  _processMessage(rawMessage) {
    const message = JSON.parse(rawMessage.data);
    const actionHandlers = {
      error: this._handleError.bind(this),
      success: this._handleCallSuccess.bind(this),
      pong: this._handlePong.bind(this),
      authenticated: this._handleAuthenticated.bind(this),
      joined: this._handleJoined.bind(this),
      join: this._handleJoinError.bind(this),
      'ot:delta': this._handleOtDelta.bind(this)
    };

    if (actionHandlers[message[0]] === undefined) {
      this.emit('message', message);
    } else {
      // drop message if request has already timed out client-side
      if (this._timeoutedRequests[message[1]]) {
        delete this._timeoutedRequests[message[1]];
      } else {
        actionHandlers[message[0]](message);
      }
    }

    this.emit('log', {
      direction: 'receive',
      data: rawMessage.data
    });
  } // request - response promise matching


  _createRequest(callback) {
    const id = this._nextRequestIndex++;
    const deferred = defer();
    this._openRequests[id] = {
      deferred,
      callback
    };
    return {
      id,
      promise: deferred.promise
    };
  }

  _closeChannel(channelName) {
    const channel = this._otChannels[channelName];
    delete this._otChannels[channelName];

    if (channel.buffer && channel.buffer.promises) {
      for (const promise of channel.buffer.promises) {
        promise.reject('closed');
      }
    }
  }

  _popPendingRequest(id) {
    const req = this._openRequests[id];
    this._openRequests[id] = undefined;
    return req;
  }

  _handleError(message) {
    const req = this._popPendingRequest(message[1]);

    if (req === null || req === undefined) {
      this.emit('error', message[message.length - 1]);
    } else {
      req.deferred.reject(new APIError(message[2]));
    }
  }

  _handleCallSuccess(message) {
    const req = this._popPendingRequest(message[1]);

    if (req === null || req === undefined) {
      this.emit('warning', `no saved request with id: ${message[1]}`);
    } else {
      if (req.callback) {
        req.callback(message[2], req.deferred);
      } else {
        req.deferred.resolve(message[2]);
      }
    }
  }

  _handlePong(message) {
    this.emit('pong');
    this._pingState.latestPong = Date.now();
  }

  _handleAuthenticated() {
    clearTimeout(this._authenticationTimeout);
    this._authenticationTimeout = null;
    this._joinTimeout = setTimeout(() => {
      this._handlePingTimeout();
    }, this._config.joinTimeout);
  }

  _handleJoined(message) {
    clearTimeout(this._joinTimeout);
    this._joinTimeout = null;

    if (message[1].channels) {
      // initialize channel revision
      const newChannels = Object.keys(message[1].channels);
      const closedChannels = Object.keys(this._otChannels).filter(key => !newChannels.includes(key));

      for (const channelName of newChannels) {
        this._otChannels[channelName] = {
          deltaInFlight: null,
          buffer: null,
          rev: message[1].channels[channelName].last_revision
        };
      }

      for (const channelName of closedChannels) {
        this._closeChannel(channelName);
      }
    }

    const viewToResend = this._currentView;
    this._currentView = null;
    this.emit('joined', message[1]); // if _currentView is set again, sendView() was called from `joined` handler. Discard.

    if (viewToResend && !this._currentView) this._send(viewToResend); // start pinging

    const socket = this._socket;
    setTimeout(() => {
      if (socket === this._socket) this._ping(socket);
    }, this._config.pingInterval);
  }

  _handleJoinError(message) {
    clearTimeout(this._joinTimeout);
    this._joinTimeout = null;
    this.emit('error', message[1].error);
  }

  _handleOtDelta(message) {
    const channelName = message[1];
    const data = message[2];
    let delta = new Delta(data.delta);
    let channel = this._otChannels[channelName];

    if (!channel) {
      // somebody else started a channel
      channel = this._otChannels[channelName] = {
        deltaInFlight: null,
        buffer: null,
        rev: 0
      };
    }

    channel.rev = data.rev;

    if (channel.deltaInFlight) {
      delta = channel.deltaInFlight.transform(delta);
    }

    if (channel.buffer) {
      const reveicedDelta = delta;
      delta = channel.buffer.delta.transform(delta);
      channel.buffer.delta = reveicedDelta.transform(channel.buffer.delta, true);
    }

    return this.emit('ot:delta', channelName, delta);
  }

}

exports.APIError = APIError;
exports.BASE_TYPES = BASE_TYPES;
exports.Delta = Delta;
exports.DeltaString = DeltaString;
exports.OpIterator = OpIterator;
exports.QuidditchClient = QuidditchClient;
exports.SUBTYPES = SUBTYPES;
exports.cloneOps = clone;
exports.setSubtypes = setSubtypes;
